@@ -67,6 +67,10 @@ namespace lsvpd
 	/* NVME f1h log page VPD size */
 	#define NVME_VPD_INFO_SIZE		1024
 
+	/* NVMe-MI Receive Opcode */
+	#define NVME_MI_CMD_RECEIVE     0x1E
+	#define NVME_MI_VPD_DATA_LEN    256
+
 	/**
 	 * SysFSTreeCollector contains the logic for device discovery and VPD
 	 * retrieval from /sys and sg_utils.
@@ -125,6 +129,7 @@ namespace lsvpd
 			// nvme specific
 		        int load_nvme_templates(const string& filename);
 			int interpretNVMEf1hLogPage(Component *fillMe, char *data);
+			int interpretNVMEMiLog(Component *fillMe, char *data);
 
 			// scsi specific
 			int load_scsi_templates(const string&);
@@ -139,6 +144,8 @@ namespace lsvpd
 			void process_template(Component *fillMe, string *deviceType,
 									char *data, int dataSize, string *format,
 									int pageCode);
+			void fillSpyreVpd(Component* fillMe);
+			string read11S(unsigned char* bar0_ptr);
 			int collectNvmeVpd(Component *fillMe, int device_fd);
 
 			int collectVpd(Component *fillMe, int device_fd, bool limitSCSISize);
@@ -204,6 +211,17 @@ namespace lsvpd
 			 */
 			void fillPCIDev( Component* fillMe, const string& sysDir );
 
+			/** Set custom VPD fields for specific devices based on
+			 * device/vendor IDs.
+			 * @param fillMe: Component to populate with custom VPD data
+			 * @param vendorId: PCI vendor ID
+			 * @param deviceId: PCI device ID
+			 * @param subVendorId: PCI subsystem vendor ID
+			 * @param subDeviceId: PCI subsystem device ID
+			 */
+			void setCustomVPDFields(Component* fillMe, int vendorId, int deviceId,
+					int subVendorId, int subDeviceId);
+
 			/**
 			 * Fill the Manufacturer name and Product description if the
 			 * vendor and product ids are available under /sys.
