@@ -35,7 +35,10 @@
 #include <libvpd-2/debug.hpp>
 #include <libvpd-2/logger.hpp>
 #include <libvpd-2/lsvpd.hpp>
+#include <libvpd-2/vpddbenv.hpp>
 
+#include <linux/vfio.h>
+#include <sys/mman.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/utsname.h>
@@ -61,10 +64,20 @@
 #include <fstream>
 #include <sstream>
 #include <iomanip>
+#include <algorithm>
+
+#define MNIMI_INST 0x080280
+#define ECID1 0x080588
+#define MNIMI_DATA 0x0802C0
+
+std::map<std::string, bool> g_deviceAccessible;
 
 extern int errno;
 
 using namespace std;
+using namespace lsvpd;
+
+extern VpdDbEnv *spyreDb;
 
 namespace lsvpd
 {
@@ -1407,6 +1420,290 @@ ERROR:
 		return 0;
 	}
 
+	void SysFSTreeCollector::fillSpyreVpd(Component* fillMe)
+	{
+		string path;
+		Logger l;
+		char device_id[16] = {0};
+		int container_fd = -1, group_fd = -1, device_fd = -1;
+		int iommu_type = -1;
+		char path_buf[256] = {0}, group_path[256] = {0};
+		char* group_name = nullptr;
+		ssize_t len = 0;
+		uint64_t eeprom_data = 0;
+		struct vfio_region_info reg = {0};
+		void* bar0_mem = nullptr;
+		unsigned char* bar0_ptr = nullptr;
+		ostringstream ss;
+		uint8_t eeprom_res_manu = 0;
+		uint8_t boot_version = 0;
+		string boot_version_str;
+		string partNumber;
+
+		/* Read device ID */
+		path = fillMe->getID() + "/device";
+		ifstream device_stream(path.c_str());
+		if (!device_stream)
+			return;
+
+		device_stream.getline(device_id, sizeof(device_id));
+		device_stream.close();
+
+		/* Check if the device is Spyre (device ID 0x06a7 or 0x06a8) */
+		if (strcmp(device_id, "0x06a7") != 0 && strcmp(device_id, "0x06a8") != 0)
+			return;
+
+		l.log("Confirmed Spyre device " + fillMe->getID() + " with ID: " + string(device_id), LOG_INFO);
+
+		/* Open VFIO container */
+		container_fd = open("/dev/vfio/vfio", O_RDWR);
+		if (container_fd < 0) {
+			l.log("Failed to open VFIO container /dev/vfio/vfio for " + fillMe->getID() +
+					", errno: " + to_string(errno) + " (" + strerror(errno) + ")", LOG_ERR);
+			return;
+		}
+		l.log("Successfully opened VFIO container for " + fillMe->getID(), LOG_DEBUG);
+
+		/* Check for IOMMU Support */
+		if (ioctl(container_fd, VFIO_CHECK_EXTENSION, VFIO_TYPE1_IOMMU) == 1) {
+			iommu_type = VFIO_TYPE1_IOMMU;
+			l.log("Using VFIO_TYPE1_IOMMU for " + fillMe->getID(), LOG_DEBUG);
+		} else if (ioctl(container_fd, VFIO_CHECK_EXTENSION, VFIO_SPAPR_TCE_IOMMU) == 1) {
+			iommu_type = VFIO_SPAPR_TCE_IOMMU;
+			l.log("Using VFIO_SPAPR_TCE_IOMMU for " + fillMe->getID(), LOG_DEBUG);
+		}
+		else {
+			l.log("No supported IOMMU type found for " + fillMe->getID(), LOG_ERR);
+			close(container_fd);
+			return;
+		}
+
+		/* Get IOMMU group */
+		snprintf(path_buf, sizeof(path_buf), "%s/iommu_group", fillMe->getID().c_str());
+		len = readlink(path_buf, group_path, sizeof(group_path) - 1);
+		if (len < 0) {
+			l.log("Failed to read IOMMU group symlink " + string(path_buf) + " for " + fillMe->getID() +
+					", errno: " + to_string(errno) + " (" + strerror(errno) + ")", LOG_ERR);
+			close(container_fd);
+			return;
+		}
+
+		group_path[len] = '\0';
+		group_name = strrchr(group_path, '/');
+		if (!group_name) {
+			l.log("Failed to extract group name from path: " + string(group_path) + " for " + fillMe->getID(), LOG_ERR);
+			close(container_fd);
+			return;
+		}
+		group_name++;
+
+		l.log("Found IOMMU group: " + string(group_name) + " for " + fillMe->getID(), LOG_DEBUG);
+
+		/* Open the VFIO Group */
+		snprintf(path_buf, sizeof(path_buf), "/dev/vfio/%s", group_name);
+		group_fd = open(path_buf, O_RDWR);
+
+		g_deviceAccessible[fillMe->getID()] = true;
+		if (group_fd < 0) {
+
+			Logger l;
+			l.log("Failed to open VFIO group " + string(path_buf) + " for " + fillMe->getID(), LOG_ERR);
+			g_deviceAccessible[fillMe->getID()] = false;
+
+			if (spyreDb != nullptr) {
+				l.log("Attempting to use cached data from spyreDb for " + fillMe->getID(), LOG_INFO);
+
+				Component* spyreComp = spyreDb->fetch(fillMe->getID());
+				if (spyreComp != nullptr) {
+					l.log("Found cached component data for " + fillMe->getID(), LOG_DEBUG);
+
+					if (!spyreComp->mManufacturer.dataValue.empty()) {
+						fillMe->mManufacturer.setValue(spyreComp->mManufacturer.getValue(), 80, __FILE__, __LINE__);
+					}
+					if (!spyreComp->mModel.dataValue.empty()) {
+						fillMe->mModel.setValue(spyreComp->mModel.getValue(), 80, __FILE__, __LINE__);
+					}
+					if (!spyreComp->mEngChangeLevel.dataValue.empty()) {
+						fillMe->mEngChangeLevel.setValue(spyreComp->mEngChangeLevel.getValue(), 80, __FILE__, __LINE__);
+					}
+					if (!spyreComp->mSerialNumber.dataValue.empty()) {
+						fillMe->mSerialNumber.setValue(spyreComp->mSerialNumber.getValue(), 80, __FILE__, __LINE__);
+					}
+					if (!spyreComp->mPartNumber.dataValue.empty()) {
+						fillMe->mPartNumber.setValue(spyreComp->mPartNumber.getValue(), 80, __FILE__, __LINE__);
+					}
+					if (!spyreComp->mPhysicalLocation.dataValue.empty()) {
+						fillMe->mPhysicalLocation.setValue(spyreComp->mPhysicalLocation.getValue(), 80, __FILE__, __LINE__);
+					}
+					if (!spyreComp->mDescription.dataValue.empty()) {
+						fillMe->mDescription.setValue(spyreComp->mDescription.getValue(), 80, __FILE__, __LINE__);
+					}
+					if (!spyreComp->mCDField.dataValue.empty()) {
+						fillMe->mCDField.setValue(spyreComp->mCDField.getValue(), 80, __FILE__, __LINE__);
+					}
+					if (!spyreComp->mFirmwareLevel.dataValue.empty()) {
+						fillMe->mFirmwareLevel.setValue(spyreComp->mFirmwareLevel.getValue(), 80, __FILE__, __LINE__);
+					}
+
+				delete spyreComp;
+				l.log("Successfully populated component data from cache for " + fillMe->getID(), LOG_INFO);
+				} else {
+					l.log("No cached component data found in spyreDb for " + fillMe->getID(), LOG_WARNING);
+				}
+			}
+
+			close(group_fd);
+			close(container_fd);
+			return;
+		}
+
+		/* Check group status */
+		struct vfio_group_status group_status = {.argsz = sizeof(group_status)};
+		if (ioctl(group_fd, VFIO_GROUP_GET_STATUS, &group_status) < 0 ||
+				!(group_status.flags & VFIO_GROUP_FLAGS_VIABLE)) {
+			l.log("Failed to get VFIO group status for " + fillMe->getID() +
+					", errno: " + to_string(errno) + " (" + strerror(errno) + ")", LOG_ERR);
+			close(group_fd);
+			close(container_fd);
+			return;
+		}
+
+		/* Add group to container */
+		if (ioctl(group_fd, VFIO_GROUP_SET_CONTAINER, &container_fd) < 0) {
+			l.log("Failed to add VFIO group to container for " + fillMe->getID() +
+					", errno: " + to_string(errno) + " (" + strerror(errno) + ")", LOG_ERR);
+			close(group_fd);
+			close(container_fd);
+			return;
+		}
+		l.log("Successfully added VFIO group to container for " + fillMe->getID(), LOG_DEBUG);
+
+		/* Set IOMMU type */
+		if (iommu_type != -1 && ioctl(container_fd, VFIO_SET_IOMMU, iommu_type) < 0) {
+			l.log("Failed to set IOMMU type " + to_string(iommu_type) + " for " + fillMe->getID() +
+					", errno: " + to_string(errno) + " (" + strerror(errno) + ")", LOG_ERR);
+			close(group_fd);
+			close(container_fd);
+			return;
+		}
+
+		l.log("Successfully set IOMMU type for " + fillMe->getID(), LOG_DEBUG);
+
+		device_fd = ioctl(group_fd, VFIO_GROUP_GET_DEVICE_FD,
+				fillMe->getID().substr(fillMe->getID().rfind("/") + 1).c_str());
+
+		if (device_fd < 0) {
+			l.log("Failed to get device FD for " + fillMe->getID() +
+					", errno: " + to_string(errno) + " (" + strerror(errno) + ")", LOG_ERR);
+			close(group_fd);
+			close(container_fd);
+			return;
+		}
+
+		l.log("Successfully obtained device FD for " + fillMe->getID(), LOG_DEBUG);
+
+		/* Configure BAR0 region info */
+		reg.argsz = sizeof(reg);
+		reg.index = VFIO_PCI_BAR0_REGION_INDEX;
+
+		/* Get BAR0 Info */
+		if (ioctl(device_fd, VFIO_DEVICE_GET_REGION_INFO, &reg) < 0) {
+			l.log("Failed to get BAR0 region info for " + fillMe->getID() +
+					", errno: " + to_string(errno) + " (" + strerror(errno) + ")", LOG_ERR);
+			close(device_fd);
+			close(group_fd);
+			close(container_fd);
+			return;
+		}
+
+		/* Map BAR0 */
+		bar0_mem = mmap(NULL, reg.size, PROT_READ | PROT_WRITE, MAP_SHARED, device_fd, reg.offset);
+		if (bar0_mem == MAP_FAILED) {
+			l.log("Failed to map BAR0 memory for " + fillMe->getID() +
+					", errno: " + to_string(errno) + " (" + strerror(errno) + ")", LOG_ERR);
+			close(device_fd);
+			close(group_fd);
+			close(container_fd);
+			return;
+		}
+
+		l.log("Successfully mapped BAR0 memory for " + fillMe->getID(), LOG_DEBUG);
+
+		/* Read data from mapped memory */
+		bar0_ptr = (unsigned char*)bar0_mem;
+
+		*((uint64_t*)(bar0_ptr + MNIMI_INST)) = 0xa101f801200b1467;
+		usleep(50000);
+
+		eeprom_data = *((uint64_t*)(bar0_ptr + MNIMI_DATA));
+
+		/* Set Manufacturer based on EEPROM data */
+		eeprom_res_manu = eeprom_data & 0xFF;
+		if (eeprom_res_manu == 0xff || eeprom_res_manu == 0x0) {
+			fillMe->mManufacturer.setValue("Samsung", 80, __FILE__, __LINE__);
+		} else if (eeprom_res_manu == 0x1) {
+			fillMe->mManufacturer.setValue("Micron", 80, __FILE__, __LINE__);
+		}
+
+		/* Set Firmware Level */
+		boot_version = (eeprom_data >> 56) & 0xFF;
+		fillMe->mFirmwareLevel.setValue(to_string(boot_version), 80, __FILE__, __LINE__);
+
+		string eeprom11s_sn = read11S(bar0_ptr);
+		if (eeprom11s_sn.length() >= 7) {
+			partNumber = eeprom11s_sn.substr(0, 7);
+			fillMe->mPartNumber.setValue(partNumber, 100, __FILE__, __LINE__);
+		}
+
+		fillMe->mEngChangeLevel.setValue(eeprom11s_sn.substr(10, 1), 100, __FILE__, __LINE__);
+		fillMe->mSerialNumber.setValue(eeprom11s_sn.substr(7), 100, __FILE__, __LINE__);
+
+                /* Clean up memory mapping */
+                if (munmap(bar0_mem, reg.size) < 0) {
+			l.log("Warning: Failed to unmap BAR0 memory for " + fillMe->getID() +
+					", errno: " + to_string(errno) + " (" + strerror(errno) + ")", LOG_WARNING);
+		} else {
+			l.log("Successfully unmapped BAR0 memory for " + fillMe->getID(), LOG_DEBUG);
+		}
+
+		/* Clean up file descriptors */
+		close(device_fd);
+		close(group_fd);
+		close(container_fd);
+
+		l.log("Successfully completed fillSpyreVpd for " + fillMe->getID(), LOG_INFO);
+	}
+
+	string SysFSTreeCollector::read11S(unsigned char* bar0_ptr)
+	{
+		string result;
+		char buffer[8];
+
+		*((uint64_t*)(bar0_ptr + MNIMI_INST)) = 0xa101e001200b1467;
+		usleep(50000);
+		uint64_t eeprom11s_0 = *((uint64_t*)(bar0_ptr + MNIMI_DATA));
+
+		*((uint64_t*)(bar0_ptr + MNIMI_INST)) = 0xa101e801200b1467;
+		usleep(50000);
+		uint64_t eeprom11s_1 = *((uint64_t*)(bar0_ptr + MNIMI_DATA));
+
+		*((uint64_t*)(bar0_ptr + MNIMI_INST)) = 0xa101f001200b1467;
+		usleep(50000);
+		uint64_t eeprom11s_2 = *((uint64_t*)(bar0_ptr + MNIMI_DATA)) & 0xFFFFFF0000000000;
+
+		uint64_t values[3] = {eeprom11s_0, eeprom11s_1, eeprom11s_2};
+
+		for (int val_idx = 0; val_idx < 3; val_idx++) {
+			for (int byte_idx = 0; byte_idx < 8; byte_idx++) {
+				buffer[byte_idx] = (values[val_idx] >> (56 - byte_idx * 8)) & 0xFF;
+			}
+			result.append(buffer, 8);
+		}
+
+		result.erase(remove(result.begin(), result.end(), '\0'), result.end());
+		return result;
+	}
+
 	void SysFSTreeCollector::fillPciNvmeVpd( Component* fillMe )
 	{
 		int device_fd;
@@ -1423,6 +1720,76 @@ ERROR:
 		return;
 	}
 
+	void SysFSTreeCollector::setCustomVPDFields(Component* fillMe,
+			int vendorId, int deviceId,
+			int subVendorId, int subDeviceId)
+	{
+		int sysPciId = (vendorId << 16) | deviceId;
+		int subSysPciId = (subVendorId << 16) | subDeviceId;
+
+		switch(sysPciId) {
+			case 0x101406A7:
+			case 0x101406A8:
+				fillMe->mFeatureCode.setValue("ECSE", 95, __FILE__, __LINE__);
+				fillMe->addDeviceSpecific( "CC", "Customer Card ID Number", "2E1F", 95);
+				break;
+			case 0x15b31021: // Nvidia ConnectX-7 PF
+				switch (subSysPciId) {
+					case 0x15b30030:  // CX7 Fractal PF
+						fillMe->mPartNumber.setValue("03JP126", 95, __FILE__, __LINE__);
+						fillMe->mEngChangeLevel.setValue("P45996", 95, __FILE__, __LINE__);
+						fillMe->mFRU.setValue("03JP128", 95, __FILE__, __LINE__);
+						fillMe->mFeatureCode.setValue("EAPC / EAPD", 95, __FILE__, __LINE__);
+						fillMe->addDeviceSpecific("CE", "CCIN Extension", "EC26", 95);
+                        fillMe->mDescription.setValue(fillMe->mDescription.getValue() +
+                            " PCIe4 4-Port 25Gb/10Gb RoCE SFP28 Adapter", 95, __FILE__, __LINE__);
+						break;
+
+					case 0x15b30022:  // CX7 Onion Creek PF
+						fillMe->mPartNumber.setValue("03JP134", 95, __FILE__, __LINE__);
+						fillMe->mEngChangeLevel.setValue("P45997", 95, __FILE__, __LINE__);
+						fillMe->mFRU.setValue("03JP136", 95, __FILE__, __LINE__);
+						fillMe->mFeatureCode.setValue("EAPE / EAPF", 95, __FILE__, __LINE__);
+						fillMe->addDeviceSpecific("CE", "CCIN Extension", "EC2E", 95);
+                        fillMe->mDescription.setValue(fillMe->mDescription.getValue() +
+                            " PCIe5 2-Port 100GbE RoCE QSFP56 Adapter", 95, __FILE__, __LINE__);
+						break;
+
+					default:
+						break;
+				}
+				break;
+
+			case 0x15b3101e:  // NVIDIA ConnectX-7 SR-IOV (Vendor: 15b3, Device: 101e)
+				switch (subSysPciId) {
+					case 0x15b30030:  // CX7 Fractal SR-IOV
+						fillMe->mPartNumber.setValue("03JP126", 95, __FILE__, __LINE__);
+						fillMe->mEngChangeLevel.setValue("P45996", 95, __FILE__, __LINE__);
+						fillMe->mFRU.setValue("03JP128", 95, __FILE__, __LINE__);
+						fillMe->mFeatureCode.setValue("EAPC / EAPD", 95, __FILE__, __LINE__);
+						fillMe->addDeviceSpecific("CE", "CCIN Extension", "EC26", 95);
+                        fillMe->mDescription.setValue(fillMe->mDescription.getValue() +
+                            " PCIe4 4-Port 25Gb/10Gb RoCE SFP28 Adapter", 95, __FILE__, __LINE__);
+						break;
+
+					case 0x15b30022:  // CX7 Onion Creek SR-IOV
+						fillMe->mPartNumber.setValue("03JP134", 95, __FILE__, __LINE__);
+						fillMe->mEngChangeLevel.setValue("P45997", 95, __FILE__, __LINE__);
+						fillMe->mFRU.setValue("03JP136", 95, __FILE__, __LINE__);
+						fillMe->mFeatureCode.setValue("EAPE / EAPF", 95, __FILE__, __LINE__);
+						fillMe->addDeviceSpecific("CE", "CCIN Extension", "EC2E", 95);
+                        fillMe->mDescription.setValue(fillMe->mDescription.getValue() +
+                            " PCIe5 2-Port 100GbE RoCE QSFP56 Adapter", 95, __FILE__, __LINE__);
+						break;
+
+					default:
+						break;
+				}
+				break;
+			}
+	}
+
+
 	/* Parse VPD file */
 	void SysFSTreeCollector::fillPciDevVpd( Component* fillMe )
 	{
@@ -1490,7 +1857,6 @@ ERROR:
 						fillMe->mManufacturer.setValue( val,
 							50, __FILE__, __LINE__ );
 
-					os << hex << setw( 4 ) << setfill( '0' ) << manID;
 				}
 			}
 			else
@@ -1500,7 +1866,6 @@ ERROR:
 					fillMe->mManufacturer.setValue( val, 50,
 							__FILE__, __LINE__ );
 
-				os << hex << setw( 4 ) << setfill( '0' ) << subMan;
 			}
 
 			// Fill Device Model
@@ -1509,7 +1874,6 @@ ERROR:
 			{
 				if( manID != UNKNOWN_ID )
 				{
-					os << hex << setw( 4 ) << setfill( '0' )<< devID;
 					val = mPciTable->getName( manID, devID );
 					if( val != "" )
 						fillMe->mModel.setValue( val, 80,
@@ -1518,7 +1882,6 @@ ERROR:
 			}
 			else
 			{
-				os << hex << setw( 4 ) << setfill( '0' )<< subID;
 				val = mPciTable->getName( manID, devID, subID );
 				if( val != "" )
 					fillMe->mModel.setValue( val, 80, __FILE__, __LINE__ );
@@ -1527,7 +1890,12 @@ ERROR:
 					fillMe->mDescription.setValue( val, 80, __FILE__, __LINE__ );
 			}
 
-			if( os.str( ) != "ffffffff" )
+			os << "(" << hex << setw(4) << setfill('0') << manID << ","
+				<< hex << setw(4) << setfill('0') << devID << "), ("
+				<< hex << setw(4) << setfill('0') << subMan << ","
+				<< hex << setw(4) << setfill('0') << subID << ")";
+
+			if( os.str( ) != "(ffff,ffff), (ffff,ffff)" )
 				fillMe->mCDField.setValue( os.str( ), 100, __FILE__, __LINE__ );
 		}
 
@@ -1536,9 +1904,15 @@ ERROR:
 		/* Fill PCI device VPD info */
 		fillPciDevVpd(fillMe);
 
+		/* Set custom VPD fields for specific devices */
+		setCustomVPDFields(fillMe, manID, devID, subMan, subID);
+
 		/* Fill NVME device VPD info using f1h log page */
 		fillPciNvmeVpd(fillMe);
 
+		/* Fill Spyre information */
+		fillSpyreVpd(fillMe);
+
 		// Read the pci config file for Device Specific (YC)
 		os.str( "" );
 		os << fillMe->sysFsNode.dataValue << "/config";
