@@ -332,31 +332,12 @@ void printVPD( Component* root )
 			}
 
 			string val = string( root->getCD( ) );
-
-			if( val != "" )
-			{
-				char s1, s2;
-				s1 = val[ 0 ];
-				s2 = val[ 1 ];
-				val[ 0 ] = val[ 2 ];
-				val[ 1 ] = val[ 3 ];
-				val[ 2 ] = s1;
-				val[ 3 ] = s2;
-
-				s1 = val[ 4 ];
-				s2 = val[ 5 ];
-				val[ 4 ] = val[ 6 ];
-				val[ 5 ] = val[ 7 ];
-				val[ 6 ] = s1;
-				val[ 7 ] = s2;
-			}
-
 			ostringstream os;
 			os << root->getDescription( );
 
 			if( val != "" )
 			{
-				os << " (" << val << ")";
+				os << " " << val;
 			}
 
 			if(	val.length( ) + root->getDescription( ).length( ) <= 35 )
@@ -418,6 +399,10 @@ void printVPD( Component* root )
 					printField( root->getSerialNumberHN( ),
 						    root->getSerialNumber( ) );
 
+                if( root->getFeatureCode( ) != "" )
+                    printField( root->getFeatureCodeHN( ),
+                            root->getFeatureCode( ) );
+
 				if( root->getFirmwareVersion( ) != "" )
 					printField( root->getFirmwareVersionHN( ),
 						    root->getFirmwareVersion( ) );
